/*
 *  <Short Description>
 *  Copyright (C) 2024  Brett Terpstra
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <blt/gfx/loader/obj_loader.h>
#include <blt/std/loader.h>
#include <blt/std/string.h>
#include <blt/std/logging.h>
#include <cctype>
#include <charconv>
#include "blt/std/assert.h"

namespace blt::gfx
{
    class char_tokenizer
    {
        private:
            std::string_view string;
            std::size_t current_pos = 0;
        public:
            explicit char_tokenizer(std::string_view view): string(view)
            {}
            
            inline char peek()
            {
                return string[current_pos];
            }
            
            inline char advance()
            {
                return string[current_pos++];
            }
            
            inline bool has_next(size_t offset = 0)
            {
                return current_pos + offset < string.size();
            }
            
            inline std::string_view read_fully()
            {
                return blt::string::trim(string.substr(current_pos));
            }
    };
    
    float get(std::string_view str)
    {
        float x;
        const auto [ptr, ec] = std::from_chars(str.data(), str.data() + str.size(), x);
        // probably not needed.
        if (ec != std::errc())
        {
//            int i;
//            const auto [ptr2, ec2] = std::from_chars(str.data(), str.data() + str.size(), i);
//            if (ec2 == std::errc())
//            {
//                x = static_cast<float>(i);
//            } else
//            {
            BLT_WARN("Unable to parse string '%s' into number!", std::string(str).c_str());
            x = 0;
//            }
        }
        return x;
    }
    
    void obj_loader::parse_extra_line(char_tokenizer& tokenizer)
    {
        char type = tokenizer.advance();
        
        if (type == 'p')
        {
            BLT_WARN("Unexpected type '%c' (not supported)", type);
            return;
        }
        
        auto elements = blt::string::split(std::string(tokenizer.read_fully()), " ");
        BLT_ASSERT(elements.size() >= 2 && "Current line doesn't have enough arguments to process!");
        float x = get(elements[0]), y = get(elements[1]);
        if (elements.size() < 3)
        {
            if (type == 't')
                uvs.push_back(uv_t{x, y});
            else
                BLT_ERROR("Unable to parse line '%s' type '%c' not recognized for arg count", std::string(tokenizer.read_fully()).c_str(), type);
        } else
        {
            float z = get(elements[2]);
            if (!handle_vertex_and_normals(x, y, z, type))
                BLT_ERROR("Unable to parse line '%s' type '%c' not recognized", std::string(tokenizer.read_fully()).c_str(), type);
        }
    }
    
    bool obj_loader::handle_vertex_and_normals(float x, float y, float z, char type)
    {
        if (std::isspace(type))
        {
            vertices.push_back(vertex_t{x, y, z});
        } else if (type == 'n')
        {
            normals.push_back(normal_t{x, y, z});
        } else
            return false;
        return true;
    }
    
    std::vector<obj_object_t> quick_load(std::string_view file)
    {
        std::vector<obj_object_t> objects;
        
        
        return objects;
    }
    
    void obj_loader::parseFile(std::string_view file)
    {
        auto lines = blt::fs::getLinesFromFile(std::string(file));
        for (const auto& line : lines)
        {
            char_tokenizer token(line);
            if (!token.has_next() || token.read_fully().empty())
                continue;
            switch (token.advance())
            {
                case '#':
                    continue;
                case 'f':
                    parse_face(token);
                    break;
                case 'v':
                    parse_extra_line(token);
                    break;
                case 'o':
                {
                    break;
                }
            }
        }
    }
    
    void obj_loader::parse_face(char_tokenizer& tokenizer)
    {
    
    }
}